from typing import Optional

import resource
import sys

from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnClassificationAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA
from autosklearn.pipeline.implementations.util import softmax
from autosklearn.util.common import check_for_bool, check_none


class LibSVM_SVC(AutoSklearnClassificationAlgorithm):
    def __init__(
        self,
        C,
        kernel,
        gamma,
        shrinking,
        tol,
        max_iter,
        class_weight=None,
        degree=3,
        coef0=0,
        random_state=None,
    ):
        self.C = C
        self.kernel = kernel
        self.degree = degree
        self.gamma = gamma
        self.coef0 = coef0
        self.shrinking = shrinking
        self.tol = tol
        self.class_weight = class_weight
        self.max_iter = max_iter
        self.random_state = random_state
        self.estimator = None

    def fit(self, X, Y):
        import sklearn.svm

        # Calculate the size of the kernel cache (in MB) for sklearn's LibSVM.
        # The cache size is calculated as 2/3 of the available memory
        # (which is calculated as the memory limit minus the used memory)
        try:
            # Retrieve memory limits imposed on the process
            soft, hard = resource.getrlimit(resource.RLIMIT_AS)

            if soft > 0:
                # Convert limit to units of megabytes
                soft /= 1024 * 1024

                # Retrieve memory used by this process
                maxrss = resource.getrusage(resource.RUSAGE_SELF)[2] / 1024

                # In MacOS, the MaxRSS output of resource.getrusage in bytes;
                # on other platforms, it's in kilobytes
                if sys.platform == "darwin":
                    maxrss = maxrss / 1024

                cache_size = (soft - maxrss) / 1.5

                if cache_size < 0:
                    cache_size = 200
            else:
                cache_size = 200
        except Exception:
            cache_size = 200

        self.C = float(self.C)
        if self.degree is None:
            self.degree = 3
        else:
            self.degree = int(self.degree)
        if self.gamma is None:
            self.gamma = 0.0
        else:
            self.gamma = float(self.gamma)
        if self.coef0 is None:
            self.coef0 = 0.0
        else:
            self.coef0 = float(self.coef0)
        self.tol = float(self.tol)
        self.max_iter = float(self.max_iter)

        self.shrinking = check_for_bool(self.shrinking)

        if check_none(self.class_weight):
            self.class_weight = None

        self.estimator = sklearn.svm.SVC(
            C=self.C,
            kernel=self.kernel,
            degree=self.degree,
            gamma=self.gamma,
            coef0=self.coef0,
            shrinking=self.shrinking,
            tol=self.tol,
            class_weight=self.class_weight,
            max_iter=self.max_iter,
            random_state=self.random_state,
            cache_size=cache_size,
            decision_function_shape="ovr",
        )
        self.estimator.fit(X, Y)
        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    def predict_proba(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        decision = self.estimator.decision_function(X)
        return softmax(decision)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "LibSVM-SVC",
            "name": "LibSVM Support Vector Classification",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        C = UniformFloatHyperparameter("C", 0.03125, 32768, log=True, default_value=1.0)
        # No linear kernel here, because we have liblinear
        kernel = CategoricalHyperparameter(
            name="kernel", choices=["rbf", "poly", "sigmoid"], default_value="rbf"
        )
        degree = UniformIntegerHyperparameter("degree", 2, 5, default_value=3)
        gamma = UniformFloatHyperparameter(
            "gamma", 3.0517578125e-05, 8, log=True, default_value=0.1
        )
        # TODO this is totally ad-hoc
        coef0 = UniformFloatHyperparameter("coef0", -1, 1, default_value=0)
        # probability is no hyperparameter, but an argument to the SVM algo
        shrinking = CategoricalHyperparameter(
            "shrinking", ["True", "False"], default_value="True"
        )
        tol = UniformFloatHyperparameter(
            "tol", 1e-5, 1e-1, default_value=1e-3, log=True
        )
        # cache size is not a hyperparameter, but an argument to the program!
        max_iter = UnParametrizedHyperparameter("max_iter", -1)

        cs = ConfigurationSpace()
        cs.add_hyperparameters(
            [C, kernel, degree, gamma, coef0, shrinking, tol, max_iter]
        )

        degree_depends_on_poly = EqualsCondition(degree, kernel, "poly")
        coef0_condition = InCondition(coef0, kernel, ["poly", "sigmoid"])
        cs.add_condition(degree_depends_on_poly)
        cs.add_condition(coef0_condition)

        return cs



from typing import Iterator, List, Tuple, Type

import warnings
from contextlib import contextmanager

from sklearn.exceptions import ConvergenceWarning

regressor_warnings = [
    (
        UserWarning,
        (  # From QuantileTransformer
            r"n_quantiles \(\d+\) is greater than the total number of samples \(\d+\)\."
            r" n_quantiles is set to n_samples\."
        ),
    ),
    (
        ConvergenceWarning,
        (  # From GaussianProcesses
            r"The optimal value found for dimension \d+ of parameter \w+ is close"
            r" to the specified (upper|lower) bound .*(Increasing|Decreasing) the bound"
            r" and calling fit again may find a better value."
        ),
    ),
    (
        UserWarning,
        (r"n_components is too large: it will be set to \d+"),  # From FastICA
    ),
    (
        ConvergenceWarning,
        (  # From SGD
            r"Maximum number of iteration reached before convergence\."
            r" Consider increasing max_iter to improve the fit\."
        ),
    ),
    (
        ConvergenceWarning,
        (  # From MLP
            r"Stochastic Optimizer: Maximum iterations \(\d+\) reached and the"
            r" optimization hasn't converged yet\."
        ),
    ),
]

classifier_warnings = [
    (
        UserWarning,
        (  # From QuantileTransformer
            r"n_quantiles \(\d+\) is greater than the total number of samples \(\d+\)\."
            r" n_quantiles is set to n_samples\."
        ),
    ),
    (
        UserWarning,
        (r"n_components is too large: it will be set to \d+"),  # From FastICA
    ),
    (
        ConvergenceWarning,
        (  # From Liblinear
            r"Liblinear failed to converge, increase the number of iterations\."
        ),
    ),
    (
        ConvergenceWarning,
        (  # From SGD
            r"Maximum number of iteration reached before convergence\."
            r"Consider increasing max_iter to improve the fit\."
        ),
    ),
    (
        ConvergenceWarning,
        (  # From MLP
            r"Stochastic Optimizer: Maximum iterations \(\d+\) reached and the"
            r" optimization hasn't converged yet\."
        ),
    ),
    (
        ConvergenceWarning,
        (  # From FastICA
            r"FastICA did not converge\."
            r" Consider increasing tolerance or the maximum number of iterations\."
        ),
    ),
    (
        UserWarning,
        (r"Variables are collinear"),  # From LDA (Linear Discriminant Analysis)
    ),
    (
        UserWarning,
        (
            r"Clustering metrics expects discrete values but received continuous values"
            r" for label, and multiclass values for target"
        ),
    ),
]

feature_preprocessing_warnings = [
    (
        ConvergenceWarning,
        (  # From liblinear
            r"Liblinear failed to converge, increase the number of iterations."
        ),
    )
]

ignored_warnings = (
    regressor_warnings + classifier_warnings + feature_preprocessing_warnings
)


@contextmanager
def ignore_warnings(
    to_ignore: List[Tuple[Type[Warning], str]] = ignored_warnings
) -> Iterator[None]:
    """A context manager to ignore warnings

    >>> with ignore_warnings(classifier_warnings):
    >>>     ...

    Parameters
    ----------
    to_ignore: List[Tuple[Exception, str]] = ignored_warnings
        The list of warnings to ignore, defaults to all registered warnings
    """
    with warnings.catch_warnings():
        for category, message in to_ignore:
            warnings.filterwarnings("ignore", category=category, message=message)
        yield



import numpy as np
import sklearn.datasets

from autosklearn.pipeline.components.data_preprocessing.rescaling import RescalingChoice
from autosklearn.pipeline.util import get_dataset

import unittest


class ScalingComponentTest(unittest.TestCase):
    def _test_helper(self, Preprocessor, dataset=None, make_sparse=False):
        X_train, Y_train, X_test, Y_test = get_dataset(
            dataset=dataset,
            make_sparse=make_sparse,
        )

        dataset_properties = {"sparse": make_sparse}

        original_X_train = X_train.copy()
        configuration_space = Preprocessor(
            dataset_properties
        ).get_hyperparameter_search_space(dataset_properties=dataset_properties)
        default = configuration_space.get_default_configuration()

        preprocessor = Preprocessor(dataset_properties, random_state=1)
        preprocessor.set_hyperparameters(default)
        preprocessor = preprocessor.choice
        transformer = preprocessor.fit(X_train, Y_train)
        return transformer.transform(X_train), original_X_train

    def test_boston_is_not_scaled(self):
        data = sklearn.datasets.load_boston()["data"]
        self.assertGreaterEqual(np.max(data), 100)

    def test_default_configuration(self):
        transformations = []
        for i in range(2):
            transformation, original = self._test_helper(
                RescalingChoice, dataset="boston"
            )
            # The maximum is around 1.95 for the transformed array...
            self.assertAlmostEqual(np.mean(transformation), 0, places=5)
            self.assertAlmostEqual(np.std(transformation), 1, places=5)
            self.assertFalse((original == transformation).all())
            transformations.append(transformation)
            if len(transformations) > 1:
                self.assertTrue((transformations[-1] == transformations[-2]).all())

    def test_default_configuration_with_sparse_data(self):
        preprocessing = self._test_helper(
            RescalingChoice, dataset="boston", make_sparse=True
        )
        transformation, original = preprocessing
        self.assertEqual(original.getnnz(), transformation.getnnz())
        self.assertTrue(~np.allclose(original.data, transformation.data))

    @unittest.skip("Does not work at the moment.")
    def test_preprocessing_dtype(self):
        super(ScalingComponentTest, self)._test_helper(RescalingChoice)



"""Check the internal state of the automl instances after it has been fitted"""

from pathlib import Path

from autosklearn.automl import AutoML
from autosklearn.ensemble_building.builder import CANDIDATES_FILENAME

from pytest import mark
from pytest_cases import parametrize_with_cases
from pytest_cases.filters import has_tag

import test.test_automl.cases as cases
from test.conftest import DEFAULT_SEED

# Some filters
has_ensemble = has_tag("fitted") & ~has_tag("no_ensemble")
no_ensemble = has_tag("fitted") & has_tag("no_ensemble")


@parametrize_with_cases("automl", cases=cases, has_tag=["fitted", "holdout"])
def test_holdout_loaded_models(automl: AutoML) -> None:
    """
    Parameters
    ----------
    automl : AutoML
        The fitted automl object to test

    Expects
    -------
    * The ensemble should not be empty
    * The models_ should contain the identifiers of what's in the ensemble
    * The cv_models_ attr should remain None
    """
    assert automl.ensemble_ is not None

    ensemble_identifiers = automl.ensemble_.get_selected_model_identifiers()

    assert set(automl.models_.keys()) == set(ensemble_identifiers)
    assert automl.cv_models_ is None


@parametrize_with_cases("automl", cases=cases, has_tag=["fitted", "cv"])
def test_cv_loaded_models(automl: AutoML) -> None:
    """
    Parameters
    ----------
    automl : AutoML
        The fitted automl object to test

    Expects
    -------
    * The ensemble should not be empty
    * The models_ should contain the identifiers of what's in the ensemble
    * The cv_models_ should contain the identifiers of what's in the ensemble
    """
    assert automl.ensemble_ is not None

    ensemble_identifiers = automl.ensemble_.get_selected_model_identifiers()

    assert set(automl.models_.keys()) == set(ensemble_identifiers)
    assert set(automl.cv_models_.keys()) == set(ensemble_identifiers)


@parametrize_with_cases("automl", cases=cases, has_tag=no_ensemble)
def test_no_ensemble(automl: AutoML) -> None:
    """
    Parameters
    ----------
    automl : AutoML
        A fitted automl object with ensemble size specified as 0

    Expects
    -------
    * Auto-sklearn does not load a model
    * The models_ should be of size 0
    * The cv_models_ should remain None
    """
    assert automl.ensemble_ is None
    assert len(automl.models_) == 0
    assert len(automl.cv_models_) == 0


@mark.todo
def test_datamanager_stored_contents() -> None:
    """
    Expects
    -------
    * TODO
    """
    ...


@parametrize_with_cases("automl", cases=cases, filter=has_ensemble)
def test_paths_created(automl: AutoML) -> None:
    """
    Expects
    -------
    * The given paths should exist after the automl has been run and fitted
    """
    assert automl._backend is not None

    partial = Path(automl._backend.internals_directory)
    expected = [
        partial / fixture
        for fixture in (
            "true_targets_ensemble.npy",
            f"start_time_{DEFAULT_SEED}",
            "datamanager.pkl",
            "runs",
        )
    ]

    for path in expected:
        assert path.exists()


@parametrize_with_cases("automl", cases=cases, filter=has_ensemble)
def test_paths_created_with_ensemble(automl: AutoML) -> None:
    """
    Expects
    -------
    * The given paths for an automl with an ensemble should include paths
    specific to ensemble building
    """
    assert automl._backend is not None

    partial = Path(automl._backend.internals_directory)
    expected = [
        partial / fixture
        for fixture in (
            "ensembles",
            "ensemble_history.json",
            CANDIDATES_FILENAME,
        )
    ]

    for path in expected:
        assert path.exists()


@parametrize_with_cases("automl", cases=cases, filter=has_ensemble)
def test_at_least_one_model_and_predictions(automl: AutoML) -> None:
    """
    Expects
    -------
    * There should be at least one models saved
    * Each model saved should have predictions for the ensemble
    """
    assert automl._backend is not None
    runs_dir = Path(automl._backend.get_runs_directory())

    runs = list(runs_dir.iterdir())
    assert len(runs) > 0

    at_least_one = False
    for run in runs:
        prediction_files = run.glob("predictions_ensemble*.npy")
        model_files = run.glob("*.*.model")

        if any(prediction_files):
            at_least_one = True
            assert any(model_files), "Run produced prediction but no model"

    assert at_least_one, "No runs produced predictions"


@parametrize_with_cases("automl", cases=cases, filter=has_ensemble)
def test_at_least_one_ensemble(automl: AutoML) -> None:
    """
    Expects
    -------
    * There should be at least one ensemble generated
    """
    assert automl._backend is not None
    ens_dir = Path(automl._backend.get_ensemble_dir())

    # TODO make more generic
    assert len(list(ens_dir.glob("*.ensemble"))) > 0



"""Test the refitting functionality of an automl instance"""
from typing import Callable, Union

from itertools import repeat

import numpy as np

from autosklearn.automl import AutoML
from autosklearn.data.validation import InputValidator

from pytest_cases import parametrize
from unittest.mock import Mock


@parametrize("budget_type", [None, "iterations"])
def test_shuffle_on_fail(
    budget_type: Union[None, str],
    make_automl: Callable[..., AutoML],
) -> None:
    """
    Parameters
    ----------
    budget_type : Union[None, str]
        The budget type to use

    Fixtures
    --------
    make_automl : Callable[..., AutoML]
        Factory to make an AutoML instance

    Expects
    -------
    * The automl should not be able to fit before `refit`
    * The model should be attempted to fitted `n_fails` times before successing once
     after
     * The automl should be able to fit after `refit`
    """
    n_fails = 3
    failing_model = Mock()
    failing_model.fit.side_effect = [ValueError()] * n_fails + [None]  # type: ignore
    failing_model.estimator_supports_iterative_fit.side_effect = repeat(False)

    ensemble_mock = Mock()
    ensemble_mock.get_selected_model_identifiers.return_value = [(1, 1, 50.0)]

    X = np.ones((3, 2))
    y = np.ones((3,))

    input_validator = InputValidator()
    input_validator.fit(X, y)

    auto = make_automl()
    auto.ensemble_ = ensemble_mock  # type: ignore
    auto.models_ = {(1, 1, 50.0): failing_model}
    auto._budget_type = budget_type
    auto.InputValidator = input_validator

    assert not auto._can_predict
    auto.refit(X, y)

    assert failing_model.fit.call_count == n_fails + 1
    assert auto._can_predict



import numpy as np

from autosklearn.constants import REGRESSION
from autosklearn.ensembles.singlebest_ensemble import SingleBest, SingleModelEnsemble
from autosklearn.metrics import root_mean_squared_error


def test_SingleBest(backend):

    ensemble = SingleBest(
        task_type=REGRESSION,
        metrics=[root_mean_squared_error],
        random_state=0,
        backend=backend,
    )

    # fit the whole thing
    X_data = np.random.random(size=(100, 2))
    y_true = np.full((100), 5.5)
    predictions = []
    for i in range(20, 1, -1):
        pred = np.ones((100), dtype=np.float32) * i
        predictions.append(pred)
    identifiers = [(i, i, 0.0) for i in range(20)]

    # Check that the weight isn't touched
    assert ensemble.weights_ == [1.0]

    ensemble.fit(
        base_models_predictions=predictions,
        X_data=X_data,
        true_targets=y_true,
        model_identifiers=identifiers,
        runs=[],
    )

    assert ensemble.weights_ == [1.0]
    assert ensemble.indices_ == [14]
    assert ensemble.identifiers_ == [(14, 14, 0.0)]
    assert ensemble.best_model_score_ == 0.5

    # test the prediction
    ensemble_prediction = np.random.random((1, 100, 2))
    prediction = ensemble.predict(ensemble_prediction)
    assert prediction.shape == (100, 2)

    # test string representation; it selects the 14th prediction (6)
    # and ties with the 15th (5), but then afterwards the RMSE goes up again
    assert (
        str(ensemble)
        == """SingleBest:
\tMembers: [14]
\tWeights: [1.0]
\tIdentifiers: [(14, 14, 0.0)]"""
    )

    models_with_weights = ensemble.get_models_with_weights(
        {identifier: i for i, identifier in enumerate(identifiers)}
    )
    assert models_with_weights == [(1.0, 14)]

    identifiers_with_weights = ensemble.get_identifiers_with_weights()
    assert identifiers_with_weights == [((14, 14, 0.0), 1.0)]

    selected_model_identifiers = ensemble.get_selected_model_identifiers()
    assert selected_model_identifiers == [(14, 14, 0.0)]

    best_model_score = ensemble.get_validation_performance()
    assert best_model_score == 0.5


def test_SingleModelEnsemble(backend):

    ensemble = SingleModelEnsemble(
        task_type=REGRESSION,
        metrics=[root_mean_squared_error],
        random_state=0,
        backend=backend,
        model_index=5,
    )

    # fit the whole thing
    X_data = np.random.random(size=(100, 2))
    y_true = np.full((100), 5.5)
    predictions = []
    for i in range(20, 1, -1):
        pred = np.ones((100), dtype=np.float32) * i
        predictions.append(pred)
    identifiers = [(i, i, 0.0) for i in range(20)]

    # Check that the weight isn't touched
    assert ensemble.weights_ == [1.0]

    ensemble.fit(
        base_models_predictions=predictions,
        X_data=X_data,
        true_targets=y_true,
        model_identifiers=identifiers,
        runs=[],
    )

    assert ensemble.weights_ == [1.0]
    assert ensemble.indices_ == [5]
    assert ensemble.identifiers_ == [(5, 5, 0.0)]
    assert ensemble.best_model_score_ == 9.5

    # test the prediction
    ensemble_prediction = np.random.random((1, 100, 2))
    prediction = ensemble.predict(ensemble_prediction)
    assert prediction.shape == (100, 2)

    assert (
        str(ensemble)
        == """SingleModelEnsemble:
\tMembers: [5]
\tWeights: [1.0]
\tIdentifiers: [(5, 5, 0.0)]"""
    )

    models_with_weights = ensemble.get_models_with_weights(
        {identifier: i for i, identifier in enumerate(identifiers)}
    )
    assert models_with_weights == [(1.0, 5)]

    identifiers_with_weights = ensemble.get_identifiers_with_weights()
    assert identifiers_with_weights == [((5, 5, 0.0), 1.0)]

    selected_model_identifiers = ensemble.get_selected_model_identifiers()
    assert selected_model_identifiers == [(5, 5, 0.0)]

    best_model_score = ensemble.get_validation_performance()
    assert best_model_score == 9.5



"""
For accessing a logger, please default to using
`get_named_client_logger(name, host, port)`
"""
from typing import Any, Dict, Iterator, Optional, TextIO, Type, cast

import logging
import logging.config
import logging.handlers
import multiprocessing
import os
import pickle
import random
import select
import socketserver
import struct
import threading
import warnings
from contextlib import contextmanager

import yaml


def setup_logger(
    output_dir: str,
    filename: Optional[str] = None,
    distributedlog_filename: Optional[str] = None,
    logging_config: Optional[Dict] = None,
) -> None:
    # logging_config must be a dictionary object specifying the configuration
    # for the loggers to be used in auto-sklearn.
    if logging_config is None:
        with open(os.path.join(os.path.dirname(__file__), "logging.yaml"), "r") as fh:
            logging_config = yaml.safe_load(fh)

    # Make sure we have a filename handler
    if "handlers" not in logging_config:
        logging_config["handlers"] = {}
    if "file_handler" not in logging_config["handlers"]:
        logging_config["handlers"]["file_handler"] = {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "filename": "autosklearn.log",
        }
    if "distributed_logfile" not in logging_config["handlers"]:
        # We have to create a file handler
        logging_config["handlers"]["distributed_logfile"] = {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "filename": "distributed.log",
        }

    if filename is None:
        filename = logging_config["handlers"]["file_handler"]["filename"]

    if distributedlog_filename is None:
        distributedlog_filename = logging_config["handlers"]["distributed_logfile"][
            "filename"
        ]

    # Make path absolute only if required
    # This is needed because this function might be called multiple times with the same
    # dict, and we don't want /path/path/<name>.log but rather just /path/<name>.log
    if os.path.sep not in logging_config["handlers"]["file_handler"]["filename"]:
        logging_config["handlers"]["file_handler"]["filename"] = os.path.join(
            output_dir, filename
        )
    if os.path.sep not in logging_config["handlers"]["distributed_logfile"]["filename"]:
        logging_config["handlers"]["distributed_logfile"]["filename"] = os.path.join(
            output_dir, distributedlog_filename
        )
    logging.config.dictConfig(logging_config)


def _create_logger(name: str) -> logging.Logger:
    return logging.getLogger(name)


def get_logger(name: str) -> "PickableLoggerAdapter":
    logger = PickableLoggerAdapter(name)
    return logger


class PickableLoggerAdapter:
    def __init__(self, name: str):
        self.name = name
        self.logger = _create_logger(name)

    def __getstate__(self) -> Dict[str, Any]:
        """
        Method is called when pickle dumps an object.

        Returns
        -------
        Dictionary, representing the object state to be pickled. Ignores
        the self.logger field and only returns the logger name.
        """
        return {"name": self.name}

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """
        Method is called when pickle loads an object. Retrieves the name and
        creates a logger.

        Parameters
        ----------
        state - dictionary, containing the logger name.

        """
        self.name = state["name"]
        self.logger = _create_logger(self.name)

    def debug(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.info(msg, *args, **kwargs)

    def warning(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.warning(msg, *args, **kwargs)

    def error(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.error(msg, *args, **kwargs)

    def exception(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.exception(msg, *args, **kwargs)

    def critical(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.critical(msg, *args, **kwargs)

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        self.logger.log(level, msg, *args, **kwargs)

    def isEnabledFor(self, level: int) -> bool:
        return self.logger.isEnabledFor(level)


def get_named_client_logger(
    name: str,
    host: str = "localhost",
    port: int = logging.handlers.DEFAULT_TCP_LOGGING_PORT,
) -> "PicklableClientLogger":
    logger = PicklableClientLogger(name=name, host=host, port=port)
    return logger


def _get_named_client_logger(
    name: str,
    host: str = "localhost",
    port: int = logging.handlers.DEFAULT_TCP_LOGGING_PORT,
) -> logging.Logger:
    """
    When working with a logging server, clients are expected to create a logger using
    this method. For example, the automl object will create a master that awaits
    for records sent through tcp to localhost.

    Ensemble builder will then instantiate a logger object that will submit records
    via a socket handler to the server.

    We do not need to use any format as the server will render the msg as it
    needs to.

    Parameters
    ----------
        name: (str)
            the name of the logger, used to tag the messages in the main log
        host: (str)
            Address of where the server is gonna look for messages
        port: (str)
            Port used to communicate with the server

    Returns
    -------
        local_loger: a logger object that has a socket handler
    """
    # Setup the logger configuration
    # We add client not only to identify that this is the client
    # communication part of the logger, but to make sure we have
    # a new singleton with the desired socket handlers
    local_logger = _create_logger("Client-" + name)
    local_logger.propagate = False
    local_logger.setLevel(logging.DEBUG)

    try:
        # Ignore mypy logging.handlers.SocketHandler has no attribute port
        # This is not the case clearly, yet MyPy assumes this is not the case
        # Even when using direct casting or getattr
        ports = [
            getattr(handler, "port", None) for handler in local_logger.handlers
        ]  # type: ignore[attr-defined]
    except AttributeError:
        # We do not want to log twice but adding multiple times the same
        # handler. So we check to what ports we communicate to
        # We can prevent errors with streamers not having a port with this try
        # block -- but it is a scenario that is unlikely to happen
        ports = []

    if port not in ports:
        socketHandler = logging.handlers.SocketHandler(host, port)
        local_logger.addHandler(socketHandler)

    return local_logger


class PicklableClientLogger(PickableLoggerAdapter):
    def __init__(self, name: str, host: str, port: int):
        self.name = name
        self.host = host
        self.port = port
        self.logger = _get_named_client_logger(name=name, host=host, port=port)

    def __getstate__(self) -> Dict[str, Any]:
        """
        Method is called when pickle dumps an object.

        Returns
        -------
        Dictionary, representing the object state to be pickled. Ignores
        the self.logger field and only returns the logger name.
        """
        return {
            "name": self.name,
            "host": self.host,
            "port": self.port,
        }

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """
        Method is called when pickle loads an object. Retrieves the name and
        creates a logger.

        Parameters
        ----------
        state - dictionary, containing the logger name.

        """
        self.name = state["name"]
        self.host = state["host"]
        self.port = state["port"]
        self.logger = _get_named_client_logger(
            name=self.name,
            host=self.host,
            port=self.port,
        )


class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def handle(self) -> None:
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)  # type: ignore[attr-defined]
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)  # type: ignore[attr-defined]
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))  # type: ignore[attr-defined]  # noqa: E501
            obj = self.unPickle(chunk)
            record = logging.makeLogRecord(obj)
            self.handleLogRecord(record)

    def unPickle(self, data: Any) -> Any:
        return pickle.loads(data)

    def handleLogRecord(self, record: logging.LogRecord) -> None:
        # logname is define in LogRecordSocketReceiver
        # Yet Mypy Cannot see this. This is needed so that we can
        # re-use the logging setup for autosklearn into the recieved
        # records
        if self.server.logname is not None:  # type: ignore  # noqa
            name = self.server.logname  # type: ignore  # noqa
        else:
            name = record.name
        logger = logging.getLogger(name)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)


def start_log_server(
    host: str,
    logname: str,
    event: threading.Event,
    port: multiprocessing.Value,
    filename: str,
    logging_config: Dict,
    output_dir: str,
) -> None:
    setup_logger(
        filename=filename, logging_config=logging_config, output_dir=output_dir
    )

    while True:
        # Loop until we find a valid port
        _port = random.randint(10000, 65535)
        try:
            receiver = LogRecordSocketReceiver(
                host=host,
                port=_port,
                logname=logname,
                event=event,
            )
            with port.get_lock():
                port.value = _port
            receiver.serve_until_stopped()
            break
        except OSError:
            continue


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    This class implement a entity that receives tcp messages on a given address
    For further information, please check
    https://docs.python.org/3/howto/logging-cookbook.html#configuration-server-example
    """

    allow_reuse_address = True

    def __init__(
        self,
        host: str = "localhost",
        port: int = logging.handlers.DEFAULT_TCP_LOGGING_PORT,
        handler: Type[LogRecordStreamHandler] = LogRecordStreamHandler,
        logname: Optional[str] = None,
        event: threading.Event = None,
    ):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.timeout = 1
        self.logname = logname
        self.event = event

    def serve_until_stopped(self) -> None:
        while True:
            rd, wr, ex = select.select([self.socket.fileno()], [], [], self.timeout)
            if rd:
                self.handle_request()
            if self.event is not None and self.event.is_set():
                break


@contextmanager
def warnings_to(logger: Optional[PicklableClientLogger] = None) -> Iterator[None]:
    """A context manager to catch warnings and send them to the logger

    If no logger is passed, warnings propogate as they normally would.

    Parameters
    ----------
    logger: Optional[PicklableClientLogger] = None
        The logger to send the warnings to
    """
    # Send to a logger if one exists
    if logger:

        with warnings.catch_warnings():

            def to_log(
                logger: PicklableClientLogger,
                message: str,
                category: Type[Warning],  # Weird but that's what it is
                filename: str,
                lineno: int,
                file: Optional[TextIO] = None,
                line: Optional[str] = None,
            ) -> None:
                logger.warning(f"{filename}:{lineno} {category.__name__}:{message}")

            # Mypy was complaining that logger didn't exist in `to_log` see here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#narrowing-and-inner-functions
            # we explicitly pass it in and have to force it's type with `cast`
            warnings.showwarning = lambda *args: to_log(
                cast(PicklableClientLogger, logger), *args
            )

            yield
    # Else do nothing, warnings go to wherever they would without this context
    else:
        yield



from __future__ import annotations

from typing import Callable

import random
import time
from pathlib import Path

import numpy as np

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.ensemble_building import EnsembleBuilder, Run
from autosklearn.metrics import Scorer, accuracy, make_scorer
from autosklearn.util.functional import bound, pairs

import pytest
from pytest_cases import fixture, parametrize
from unittest.mock import Mock, patch

from test.conftest import DEFAULT_SEED
from test.fixtures.metrics import acc_with_X_data


@fixture
def builder(make_ensemble_builder: Callable[..., EnsembleBuilder]) -> EnsembleBuilder:
    """A default ensemble builder"""
    return make_ensemble_builder()


@parametrize("kind", ["ensemble", "test"])
def test_targets(builder: EnsembleBuilder, kind: str) -> None:
    """
    Expects
    -------
    * Should be able to load each of the targets
    """
    assert builder.targets(kind) is not None


def test_available_runs(builder: EnsembleBuilder) -> None:
    """
    Expects
    -------
    * Should be able to read runs from the backends rundir where runs are tagged
      {seed}_{numrun}_{budget}
    """
    runsdir = Path(builder.backend.get_runs_directory())

    ids = {(0, i, 0.0) for i in range(1, 10)}
    paths = [runsdir / f"{s}_{n}_{b}" for s, n, b in ids]

    for path in paths:
        path.mkdir()

    available_runs = builder.available_runs()

    assert len(available_runs) == len(ids)
    for run_id in available_runs.keys():
        assert run_id in ids


def test_available_runs_with_bad_dir_contained(builder: EnsembleBuilder) -> None:
    """
    Expects
    -------
    * Should ignore dirs that aren't in format
    """
    runsdir = Path(builder.backend.get_runs_directory())

    ids = {(0, i, 0.0) for i in range(1, 10)}
    paths = [runsdir / f"{s}_{n}_{b}" for s, n, b in ids]

    bad_path = runsdir / "Im_a_bad_path"

    for path in paths + [bad_path]:
        path.mkdir()

    available_runs = builder.available_runs()
    assert len(available_runs) == len(paths)


def test_requires_loss_update_with_modified_runs(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * Should include runs that were modified, even if they have a loss
    """
    metrics = builder.metrics
    run_okay = [make_run(losses={m.name: 1 for m in metrics}) for _ in range(5)]
    run_modified = [
        make_run(losses={m.name: 1 for m in metrics}, modified=True) for _ in range(5)
    ]

    runs = run_okay + run_modified

    requires_update = builder.requires_loss_update(runs)

    assert set(run_modified) == set(requires_update)


def test_requires_loss_update_with_no_loss(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * Should include runs that have no loss
    """
    metrics = builder.metrics
    run_okay = [make_run(losses={m.name: 10 for m in metrics}) for _ in range(5)]
    run_no_loss = [make_run() for _ in range(5)]

    runs = run_okay + run_no_loss

    requires_update = builder.requires_loss_update(runs)

    assert set(run_no_loss) == set(requires_update)


def test_candidates_no_filters(
    builder: EnsembleBuilder, make_run: Callable[..., Run]
) -> None:
    """
    Expects
    -------
    * Should have nothing in common between candidates and discarded
    * Should not filter out any viable runs if no filters set. Here a viable run
      has a loss and ensemble predictions
    """
    metrics = builder.metrics
    dummy = make_run(dummy=True)
    runs = [make_run(losses={m.name: n for m in metrics}) for n in range(10)]

    candidates, discarded = builder.candidate_selection(
        runs,
        dummy,
        better_than_dummy=False,
        nbest=None,
    )

    assert len(set(candidates) & discarded) == 0
    assert len(candidates) == len(runs)
    assert len(discarded) == 0


def test_candidates_filters_runs_with_no_predictions(
    builder: EnsembleBuilder, make_run: Callable[..., Run]
) -> None:
    """
    Expects
    -------
    * Should have nothing in common between candidates and discarded
    * Should filter out runs with no "ensemble" predictions
    """
    metrics = builder.metrics
    dummy = make_run(dummy=True, losses={m.name: 2 for m in metrics})

    bad_runs = [make_run(predictions=None) for _ in range(5)]

    good_run = make_run(predictions="ensemble", losses={m.name: 1 for m in metrics})

    runs = bad_runs + [good_run]

    candidates, discarded = builder.candidate_selection(runs, dummy)

    assert len(set(candidates) & discarded) == 0
    assert len(candidates) == 1
    assert len(discarded) == len(bad_runs)
    assert candidates[0].pred_path("ensemble").exists()


def test_candidates_filters_runs_with_no_loss(
    builder: EnsembleBuilder, make_run: Callable[..., Run]
) -> None:
    """
    Expects
    -------
    * Should filter out runs with no loss
    """
    metrics = builder.metrics

    bad_runs = [make_run(losses=None) for _ in range(5)]
    dummy_run = make_run(dummy=True, losses={m.name: 2 for m in metrics})
    good_run = make_run(losses={m.name: 1 for m in metrics})

    runs = bad_runs + [good_run]

    candidates, discarded = builder.candidate_selection(runs, dummy_run)

    assert len(candidates) == 1
    assert len(discarded) == len(bad_runs)

    # Only candidate should be the one with a loss of 1
    assert candidates[0].losses == {m.name: 1 for m in metrics}


def test_candidates_filters_out_better_than_dummy(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * Should filter out runs worse than dummy
    """
    metrics = builder.metrics

    bad_runs = [make_run(losses={m.name: 1 for m in metrics}) for _ in range(2)]
    dummy_run = make_run(dummy=True, losses={m.name: 0 for m in metrics})
    good_runs = [make_run(losses={m.name: -1 for m in metrics}) for _ in range(3)]

    runs = bad_runs + good_runs

    candidates, discarded = builder.candidate_selection(
        runs,
        dummy_run,
        better_than_dummy=True,
    )

    assert set(candidates)

    assert len(candidates) == 3

    for run, metric in zip(candidates, metrics):
        assert run.losses[metric.name] < dummy_run.losses[metric.name]

    assert len(discarded) == 2

    for run, metric in zip(discarded, metrics):
        assert run.losses[metric.name] >= dummy_run.losses[metric.name]


def test_candidates_uses_dummy_if_no_candidates_better(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * If no run is better than a dummy run, the candidates will then consist
      of the dummy runs.
    """
    metrics = builder.metrics

    runs = [make_run(losses={m.name: 10 for m in metrics}) for _ in range(10)]
    dummies = [
        make_run(dummy=True, losses={m.name: 0 for m in metrics}) for _ in range(2)
    ]

    candidates, discarded = builder.candidate_selection(
        runs,
        dummies,
        better_than_dummy=True,
    )

    assert len(candidates) == 2
    assert all(run.is_dummy() for run in candidates)


def test_candidates_better_than_dummy_multiobjective(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * For a run to be considered a candidate, it must be better than the best dummy for
      any one of the objectives.
    """
    x = make_scorer("x", lambda: None)
    y = make_scorer("y", lambda: None)

    # The will be one best dummy per objective achieving a loss of 1
    # You can think of this as a joint filter [1, 1] over "x", "y" losses
    dummies = [
        make_run(dummy=True, losses={"x": 1, "y": 500}),
        make_run(dummy=True, losses={"x": 500, "y": 1}),
    ]

    # Clearly better on both objectives
    good_runs = [make_run(losses={"x": 0, "y": 0}) for _ in range(5)]

    # Worse on both
    badd_runs = [make_run(losses={"x": 2, "y": 2}) for _ in range(5)]

    # Better on 1 objective but worse on the other
    may1_runs = [make_run(losses={"x": 2, "y": 0}) for _ in range(5)]
    may2_runs = [make_run(losses={"x": 0, "y": 2}) for _ in range(5)]

    # Better on 1 objective but equal one the other
    may3_runs = [make_run(losses={"x": 1, "y": 0}) for _ in range(5)]
    may4_runs = [make_run(losses={"x": 0, "y": 1}) for _ in range(5)]

    # Equal on 1 objective but worse one the other
    may5_runs = [make_run(losses={"x": 1, "y": 2}) for _ in range(5)]
    may6_runs = [make_run(losses={"x": 2, "y": 1}) for _ in range(5)]

    expected_candidates = [
        *good_runs,
        *may1_runs,
        *may2_runs,
        *may3_runs,
        *may4_runs,
    ]

    expected_discarded = [
        *badd_runs,
        *may5_runs,
        *may6_runs,
    ]

    runs = expected_candidates + expected_discarded

    candidates, discarded = builder.candidate_selection(
        runs, dummies, better_than_dummy=True, metrics=[x, y]
    )

    # No duplicates
    assert len(candidates) == len(set(candidates))

    assert set(candidates) == set(expected_candidates)
    assert discarded == set(expected_discarded)


@parametrize("nbest", [0, 1, 5, 1000])
def test_candidates_nbest_int_single_objective(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
    nbest: int,
) -> None:
    """
    Expects
    -------
    * Should only select the nbest candidates
    * They should be ordered by loss
    """
    # Make sure it's single objective being tested
    assert len(builder.metrics) == 1
    metric = builder.metrics[0]

    n = 10
    expected = int(bound(nbest, bounds=(1, n)))

    dummy = make_run(dummy=True, losses={metric.name: 50_000})
    runs = [make_run(losses={metric.name: i}) for i in range(n)]
    candidates, discarded = builder.candidate_selection(runs, dummy, nbest=nbest)

    assert len(candidates) == expected

    if len(candidates) > 1:
        for a, b in pairs(candidates):
            assert a.losses[metric.name] <= b.losses[metric.name]

    # Make sure all discarded are worse than the worst candidate
    if any(discarded):
        worst_candidate_loss = candidates[-1].losses[metric.name]

        for d in discarded:
            assert worst_candidate_loss <= d.losses[metric.name]


@parametrize("nbest", [0.0, 0.25, 0.5, 1.0])
def test_candidates_nbest_float_single_objective(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
    nbest: float,
) -> None:
    """
    Expects
    -------
    * Should select nbest percentage of candidates
    * They should be ordered by loss
    """
    # Make sure it's single objective being tested
    assert len(builder.metrics) == 1
    metric = builder.metrics[0]

    n = 10
    expected = int(bound(nbest * n, bounds=(1, n)))

    dummy = make_run(dummy=True, losses={metric.name: 50_000})
    runs = [make_run(id=i, losses={metric.name: i}) for i in range(2, n + 2)]

    candidates, discarded = builder.candidate_selection(runs, dummy, nbest=nbest)

    assert len(candidates) == expected

    if len(candidates) > 1:
        for a, b in pairs(candidates):
            assert a.losses[metric.name] <= b.losses[metric.name]

    # Make sure all discarded are worse than the worst candidate
    if any(discarded):
        worst_candidate_loss = candidates[-1].losses[metric.name]

        for d in discarded:
            assert worst_candidate_loss <= d.losses[metric.name]


def test_requires_deletion_does_nothing_without_params(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * All runs should be kept
    """
    runs = [make_run() for _ in range(10)]

    keep, delete = builder.requires_deletion(
        runs,
        max_models=None,
        memory_limit=None,
    )

    assert set(runs) == set(keep)
    assert len(delete) == 0


@parametrize("max_models", [0, 1, 2, 5])
def test_requires_deletion_max_models(
    builder: EnsembleBuilder,
    max_models: int,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * Should keep exactly as many models as `max_models`
    * Should not have any in common between keep and delete
    """
    runs = [make_run() for _ in range(10)]
    keep, delete = builder.requires_deletion(runs=runs, max_models=max_models)

    assert len(keep) == max_models
    assert len(delete) == len(runs) - max_models

    assert not any(set(keep) & set(delete))


@parametrize("max_models", [0, 1, 2, 5])
def test_requires_deletion_max_models_multiobjective_no_overlap(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
    max_models: int,
) -> None:
    """
    Expects
    -------
    * When deleting runs with multiple objectives, they runs kept should be done so in
      a roundrobin fashion with respect to each metric.
    """
    # In this case, the runs have no overlap as they are round robin'ed, the order they
    # are put in is the same as the order they come out where priority shifts between
    # the three objectives
    runs = [
        make_run(id=2, losses={"x": 0, "y": 100, "z": 100}),
        make_run(id=3, losses={"x": 100, "y": 0, "z": 100}),
        make_run(id=4, losses={"x": 100, "y": 100, "z": 0}),
        #
        make_run(id=5, losses={"x": 1, "y": 100, "z": 100}),
        make_run(id=6, losses={"x": 100, "y": 1, "z": 100}),
        make_run(id=7, losses={"x": 100, "y": 100, "z": 1}),
        #
        make_run(id=8, losses={"x": 2, "y": 100, "z": 100}),
        make_run(id=9, losses={"x": 100, "y": 2, "z": 100}),
        make_run(id=10, losses={"x": 100, "y": 100, "z": 2}),
    ]

    expected_keep = list(runs[:max_models])
    expected_delete = set(runs[max_models:])

    # Dummy metrics, only used for their names
    x = make_scorer("x", lambda: None)
    y = make_scorer("y", lambda: None)
    z = make_scorer("z", lambda: None)

    # Shuffle the runs to ensure correct sorting takes place
    random.shuffle(runs)

    keep, delete = builder.requires_deletion(
        runs,
        max_models=max_models,
        metrics=[x, y, z],
    )

    assert keep == expected_keep
    assert delete == expected_delete


@parametrize("max_models", [0, 1, 2, 5, 7])
def test_requires_deletion_max_models_multiobjective_overlap(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
    max_models: int,
) -> None:
    """
    Expects
    -------
    * When deleting runs with multiple objectives, they runs kept should be done so in
      a roundrobin fashion with respect to each metric. With overlap, it should
      prioritize the first objective.
    """
    # You can read the losses as a binary counter to make sense of it
    runs = [
        make_run(id=2, losses={"x": 0, "y": 0, "z": 0}),
        make_run(id=3, losses={"x": 0, "y": 0, "z": 1}),
        make_run(id=4, losses={"x": 0, "y": 1, "z": 0}),
        make_run(id=5, losses={"x": 0, "y": 1, "z": 1}),
        make_run(id=6, losses={"x": 1, "y": 0, "z": 0}),
        make_run(id=7, losses={"x": 1, "y": 0, "z": 1}),
        make_run(id=8, losses={"x": 1, "y": 1, "z": 0}),
        make_run(id=9, losses={"x": 1, "y": 1, "z": 1}),
    ]

    # The expected order is to prefer the lowest on all objectives
    expected_order = [
        # Best at all
        make_run(id=2, losses={"x": 0, "y": 0, "z": 0}),
        # Best at 2 but prioritizing the first objective
        make_run(id=3, losses={"x": 0, "y": 0, "z": 1}),
        make_run(id=4, losses={"x": 0, "y": 1, "z": 0}),
        make_run(id=6, losses={"x": 1, "y": 0, "z": 0}),
        # Best at 1
        make_run(id=5, losses={"x": 0, "y": 1, "z": 1}),
        make_run(id=7, losses={"x": 1, "y": 0, "z": 1}),
        make_run(id=8, losses={"x": 1, "y": 1, "z": 0}),
        #
        make_run(id=9, losses={"x": 1, "y": 1, "z": 1}),
    ]

    # Dummy metrics, only used for their names
    x = make_scorer("x", lambda: None)
    y = make_scorer("y", lambda: None)

    # Shuffle the runs to ensure correct sorting takes place
    random.shuffle(runs)

    keep, delete = builder.requires_deletion(
        runs,
        max_models=max_models,
        metrics=[x, y],
    )

    assert keep == expected_order[:max_models]
    assert delete == set(expected_order[max_models:])


@parametrize("memory_limit, expected", [(0, 0), (100, 0), (200, 1), (5000, 49)])
def test_requires_memory_limit_single_objective(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
    memory_limit: int,
    expected: int,
) -> None:
    """
    Expects
    -------
    * Should keep the expected amount of models
    * The kept models should be sorted by lowest loss
    * Should not have any models in common between keep and delete
    * All models kept should be better than those deleted
    """
    # Make sure it's single objective being tested
    assert len(builder.metrics) == 1
    metric = builder.metrics[0]

    runs = [make_run(mem_usage=100, losses={metric.name: -n}) for n in range(50)]
    random.shuffle(runs)

    keep, delete = builder.requires_deletion(runs=runs, memory_limit=memory_limit)

    # The cutoff for memory is (memory_limit - largest)
    # E.g.
    #   5 models at 100 ea = 500mb usage
    #   largest = 100mb
    #   memory_limit = 400mb
    #   cutoff = memory_limit - largest  (400mb - 100mb) = 300mb
    #   We can store 300mb which means the 3 best models
    assert len(keep) == expected
    assert len(delete) == len(runs) - expected

    assert not any(set(keep) & set(delete))

    if len(keep) > 2:
        for a, b in pairs(keep):
            assert a.losses[metric.name] <= b.losses[metric.name]

    # Make sure that the best run deleted is still worse than all those kept
    best_deleted = min(r.losses[metric.name] for r in delete)
    assert not any(run.losses[metric.name] > best_deleted for run in keep)


@parametrize("kind", ["ensemble", "test"])
def test_loss_with_no_ensemble_targets(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
    kind: str,
) -> None:
    """
    Expects
    -------
    * Should give a loss of np.inf if run has no predictions of a given kind
    """
    run = make_run(predictions=None)
    X_data = builder.X_data()
    metric = builder.metrics[0]

    assert builder.loss(run, metric=metric, X_data=X_data, kind=kind) == np.inf


@parametrize("kind", ["ensemble", "test"])
def test_loss_with_targets(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
    kind: str,
) -> None:
    """
    Expects
    -------
    * Should give a loss < np.inf if the predictions exist
    """
    X_data = builder.X_data(kind)
    targets = builder.targets(kind)
    metric = builder.metrics[0]

    run = make_run(predictions={kind: targets})

    assert builder.loss(run, metric=metric, X_data=X_data, kind=kind) < np.inf


def test_delete_runs(builder: EnsembleBuilder, make_run: Callable[..., Run]) -> None:
    """
    Expects
    -------
    * Should delete runs so they can not be found again by the ensemble builder
    """
    runs = [make_run(backend=builder.backend) for _ in range(5)]
    assert all(run.dir.exists() for run in runs)

    builder.delete_runs(runs)
    assert not any(run.dir.exists() for run in runs)

    loaded = builder.available_runs()
    assert len(loaded) == 0


def test_delete_runs_does_not_delete_dummy(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * Should
    """
    backend = builder.backend
    normal_runs = [make_run(backend=backend) for _ in range(5)]
    dummy_runs = [make_run(dummy=True, seed=i, backend=backend) for i in range(2)]

    runs = normal_runs + dummy_runs
    assert all(run.dir.exists() for run in runs)

    builder.delete_runs(runs)
    assert not any(run.dir.exists() for run in normal_runs)
    assert all(dummy.dir.exists() for dummy in dummy_runs)

    loaded = builder.available_runs()
    assert set(loaded.values()) == set(dummy_runs)


def test_fit_ensemble_with_no_targets_raises(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * If no ensemble targets can be found then `fit_ensemble` should fail
    """
    # Delete the targets and then try fit ensemble
    targets_path = Path(builder.backend._get_targets_ensemble_filename())
    targets_path.unlink()

    candidates = [make_run(backend=builder.backend) for _ in range(5)]
    with pytest.raises(ValueError, match="`fit_ensemble` could not find any .*"):
        builder.fit_ensemble(
            candidates=candidates,
            runs=candidates,
        )


def test_fit_ensemble_produces_ensemble(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * Should produce an ensemble if all runs have predictions
    """
    targets = builder.targets("ensemble")
    assert targets is not None

    predictions = targets
    runs = [make_run(predictions={"ensemble": predictions}) for _ in range(10)]

    ensemble = builder.fit_ensemble(candidates=runs, runs=runs)

    assert ensemble is not None


def test_fit_with_error_gives_no_ensemble(
    builder: EnsembleBuilder,
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * A run without predictions will raise an error will cause `fit_ensemble` to fail
      as it requires all runs to have valid predictions
    """
    X_data = builder.X_data("ensemble")
    targets = builder.targets("ensemble")
    assert targets is not None

    predictions = targets

    runs = [make_run(predictions={"ensemble": predictions}) for _ in range(10)]
    bad_run = make_run(predictions=None)

    runs.append(bad_run)

    with pytest.raises(FileNotFoundError):
        builder.fit_ensemble(candidates=runs, X_data=X_data, targets=targets, runs=runs)


@parametrize("time_buffer", [1, 5])
@parametrize("duration", [10, 20])
def test_run_end_at(builder: EnsembleBuilder, time_buffer: int, duration: int) -> None:
    """
    Expects
    -------
    * The limits enforced by pynisher should account for the time_buffer and duration
      to run for + a little bit of overhead that gets rounded to a second.
    """
    with patch("pynisher.enforce_limits") as pynisher_mock:
        builder.run(
            end_at=time.time() + duration,
            iteration=1,
            time_buffer=time_buffer,
            pynisher_context="forkserver",
        )
        # The 1 comes from the small overhead in conjuction with rounding down
        expected = duration - time_buffer - 1

        # The 1 comes from the small overhead in conjuction with rounding down
        expected = duration - time_buffer - 1
        assert pynisher_mock.call_args_list[0][1]["wall_time_in_s"] == expected


def test_deletion_will_not_break_current_ensemble(
    make_backend: Callable[..., Backend],
    make_ensemble_builder: Callable[..., EnsembleBuilder],
    make_run: Callable[..., Run],
) -> None:
    """
    Expects
    -------
    * When running the builder, it's previous ensemble should not have it's runs deleted
      until a new ensemble is built.
    """
    # Make a builder with this backend and limit it to only allow 10 models on disc
    builder = make_ensemble_builder(
        max_models_on_disc=10,
        seed=DEFAULT_SEED,
    )
    metric = builder.metrics[0]

    # Stick a dummy run and 10 bad runs into the backend
    datamanager = builder.backend.load_datamanager()
    targets = datamanager.data["Y_train"]

    bad_predictions = {"ensemble": np.zeros_like(targets)}
    good_predictions = {"ensemble": targets}

    make_run(dummy=True, losses={metric.name: 10000}, backend=builder.backend)
    bad_runs = [
        make_run(backend=builder.backend, predictions=bad_predictions)
        for _ in range(10)
    ]

    ens_dir = Path(builder.backend.get_ensemble_dir())

    # Make sure there's no ensemble and run with the candidates available
    assert not ens_dir.exists()
    builder.main(time_left=100)

    # Make sure an ensemble was built
    assert ens_dir.exists()
    first_builder_contents = set(ens_dir.iterdir())

    # Create 10 new and better runs and put them in the backend
    new_runs = [
        make_run(backend=builder.backend, predictions=good_predictions)
        for _ in range(10)
    ]

    # Now we make `save_ensemble` crash so that even though we run the builder, it does
    # not manage to save the new ensemble
    with patch.object(builder.backend, "save_ensemble", side_effect=ValueError):
        try:
            builder.main(time_left=100)
        except Exception:
            pass

    # Ensure that no new ensemble was created
    second_builder_contents = set(ens_dir.iterdir())
    assert first_builder_contents == second_builder_contents

    # Now we make sure that the ensemble there still has access to all the bad models
    # that it contained from the first run, even though the second crashed.
    available_runs = builder.available_runs().values()
    for run in bad_runs + new_runs:
        assert run in available_runs

    # As a sanity check, run the builder one more time without crashing and make
    # sure the bad runs are removed with the good ones kept.
    # We remove its previous candidates so that it won't remember previous candidates
    # and will fit a new ensemble
    builder.previous_candidates_path.unlink()
    builder.main(time_left=100)
    available_runs = builder.available_runs().values()

    for run in bad_runs:
        assert run not in available_runs

    for run in new_runs:
        assert run in available_runs


@parametrize("metrics", [accuracy, acc_with_X_data, [accuracy, acc_with_X_data]])
def test_will_build_ensemble_with_different_metrics(
    make_ensemble_builder: Callable[..., EnsembleBuilder],
    make_run: Callable[..., Run],
    metrics: Scorer | list[Scorer],
) -> None:
    """
    Expects
    -------
    * Should be able to build a valid ensemble with different combinations of metrics
    * Should produce a validation score for both "ensemble" and "test" scores
    """
    if not isinstance(metrics, list):
        metrics = [metrics]

    builder = make_ensemble_builder(metrics=metrics)

    # Make some runs and stick them in the same backend as the builder
    # Dummy just has a terrible loss for all metrics
    make_run(
        dummy=True,
        losses={m.name: 1000 for m in metrics},
        backend=builder.backend,
    )

    # "Proper" runs will have the correct targets and so be better than dummy
    run_predictions = {
        "ensemble": builder.targets("ensemble"),
        "test": builder.targets("test"),
    }
    for _ in range(5):
        make_run(predictions=run_predictions, backend=builder.backend)

    history, nbest = builder.main()

    # Should only produce one step
    assert len(history) == 1
    hist = history[0]

    # Each of these two keys should be present
    for key in ["ensemble_optimization_score", "ensemble_test_score"]:
        assert key in hist

        # TODO should be updated in next PR
        #   Each of these scores should contain all the metrics
        # for metric in metrics:
        #   assert metric.name in hist[key]


@parametrize("n_least_prioritized", [1, 2, 3, 4])
@parametrize("metrics", [accuracy, acc_with_X_data, [accuracy, acc_with_X_data]])
def test_fit_ensemble_kwargs_priorities(
    make_ensemble_builder: Callable[..., EnsembleBuilder],
    make_run: Callable[..., Run],
    metrics: Scorer | list[Scorer],
    n_least_prioritized: int,
) -> None:
    """
    Expects
    -------
    * Should favour 1) function kwargs, 2) function params 3) init_kwargs 4) init_params
    """
    if not isinstance(metrics, list):
        metrics = [metrics]

    class FakeEnsembleClass:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

        def fit(*args, **kwargs) -> None:
            pass

    # We establish the priorty order and give each one of them a custom metric
    priority = ["function_kwargs", "function_params", "init_kwargs", "init_params"]

    # We reverse the priority and use the `n_least_prioritized` ones
    # with `n_least_prioritized = 3`
    #   reversed =  ["init_params", "init_kwargs", "function_params", "function_kwargs"]
    #   used =      ["init_params", "init_kwargs", "function_params"]
    #   highest =   "function_params"
    reversed_priority = list(reversed(priority))
    used = reversed_priority[:n_least_prioritized]
    highest_priority = used[-1]

    def S(name: str) -> Scorer:
        return make_scorer(name, lambda: None)

    # We now pass in all the places this arguments could be specified
    # Naming them specifically to make it more clear in setup below
    builder_metric = [S("init_params")] if "init_params" in used else None
    fit_ensemble_metric = [S("function_params")] if "function_params" in used else None

    builder_ensemble_kwargs = (
        {"metrics": [S("init_kwargs")]} if "init_kwargs" in used else None
    )
    fit_ensemble_kwargs = (
        {"metrics": [S("function_kwargs")]} if "function_kwargs" in used else None
    )

    builder = make_ensemble_builder(
        metrics=builder_metric,
        ensemble_kwargs=builder_ensemble_kwargs,
    )

    candidates = [make_run() for _ in range(5)]  # Just so something can be run

    ensemble = builder.fit_ensemble(
        metrics=fit_ensemble_metric,
        ensemble_class=FakeEnsembleClass,
        ensemble_kwargs=fit_ensemble_kwargs,
        candidates=candidates,
        runs=candidates,
    )

    # These are the final metrics passed to the ensemble builder when constructed
    passed_metrics = ensemble.kwargs["metrics"]
    metric = passed_metrics[0]

    assert metric.name == highest_priority


@parametrize("metric, should_be_loaded", [(accuracy, False), (acc_with_X_data, True)])
def test_X_data_only_loaded_when_required(
    make_ensemble_builder: Callable[..., EnsembleBuilder],
    make_run: Callable[..., Run],
    metric: Scorer,
    should_be_loaded: bool,
) -> None:
    """
    Expects
    -------
    * Should only load X_train if it's required
    * TODO should only load X_test if it's required
    """
    metrics = [metric]
    builder = make_ensemble_builder(metrics=metrics)

    # Make a dummy which is required for the whole pipeline to run
    make_run(dummy=True, losses={metric.name: 1000}, backend=builder.backend)

    # Make a run that has no losses recorded, forcing us to use the metric
    make_run(
        dummy=False,
        predictions={"ensemble": builder.targets("ensemble")},
        losses=None,
        backend=builder.backend,
    )

    ret_value = builder.X_data()
    builder.X_data = Mock(return_value=ret_value)

    builder.main()

    assert builder.X_data.called == should_be_loaded



import numpy as np


def softmax(df):
    if len(df.shape) == 1:
        df[df > 20] = 20
        df[df < -20] = -20
        ppositive = 1 / (1 + np.exp(-df))
        ppositive[ppositive > 0.999999] = 1
        ppositive[ppositive < 0.0000001] = 0
        return np.transpose(np.array((1 - ppositive, ppositive)))
    else:
        # Compute the Softmax like it is described here:
        # https://www.deeplearningbook.org/contents/numerical.html
        tmp = df - np.max(df, axis=1).reshape((-1, 1))
        tmp = np.exp(tmp)
        return tmp / np.sum(tmp, axis=1).reshape((-1, 1))


def convert_multioutput_multiclass_to_multilabel(probas):
    """Converts the model predicted probabilities to useable format.

    In some cases, models predicted_proba can output an array of shape
    (2, n_samples, n_labels) where the 2 stands for the probability of positive
    or negative. This function will convert this to an (n_samples, n_labels)
    array where the probability of a label being true is kept.

    Parameters
    ----------
    probas: array_like (1 or 2, n_samples, n_labels) or (n_samples, n_labels)
        The output of predict_proba of a classifier model

    Returns
    -------
    np.ndarray shape of (n_samples, n_labels)
        The probabilities of each label for every sample
    """
    if isinstance(probas, list):
        # probas = shape of (prob of positive/negative label, n_samples, n_labels)
        n_samples = probas[0].shape[0]
        n_labels = len(probas)
        multioutput_probas = np.ndarray((n_samples, n_labels))

        for i, label_scores in enumerate(probas):
            n_probabilities = label_scores.shape[1]

            # Label was never observed so it has a probability of 0
            if n_probabilities == 1:
                multioutput_probas[:, i] = 0

            # Label has a probability score for true or false
            elif n_probabilities == 2:
                multioutput_probas[:, i] = label_scores[:, 1]

            # In case multioutput-multiclass input was used, where we have
            # a probability for each class
            elif n_probabilities > 2:
                raise ValueError(
                    "Multioutput-Multiclass supported by "
                    "scikit-learn, but not by auto-sklearn!"
                )
            else:
                RuntimeError(f"Unkown predict_proba output={probas}")

        return multioutput_probas

    elif isinstance(probas, np.ndarray):
        if len(probas.shape) > 2:
            raise ValueError("New unsupported sklearn output!")
        else:
            return probas

    else:
        return ValueError(f"Unrecognized probas\n{type(probas)}\n{probas}")



from __future__ import annotations

from typing import Any, Dict, List, Optional, Sequence, TextIO, Tuple, Type, Union, cast

import logging
import multiprocessing
import time
import warnings

import numpy as np
from ConfigSpace import Configuration
from sklearn.base import BaseEstimator
from sklearn.dummy import DummyClassifier, DummyRegressor
from sklearn.ensemble import VotingClassifier, VotingRegressor
from smac.tae import StatusType
from threadpoolctl import threadpool_limits

import autosklearn.pipeline.classification
import autosklearn.pipeline.regression
from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.constants import (
    CLASSIFICATION_TASKS,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION_TASKS,
)
from autosklearn.data.target_validator import (
    SUPPORTED_TARGET_TYPES,
    SUPPORTED_XDATA_TYPES,
)
from autosklearn.metrics import Scorer, calculate_losses
from autosklearn.pipeline.components.base import ThirdPartyComponents, _addons
from autosklearn.pipeline.implementations.util import (
    convert_multioutput_multiclass_to_multilabel,
)
from autosklearn.util.logging_ import PicklableClientLogger, get_named_client_logger

# General TYPE definitions for numpy
TYPE_ADDITIONAL_INFO = Dict[str, Union[int, float, str, Dict, List, Tuple]]


class MyDummyClassifier(DummyClassifier):
    def __init__(
        self,
        config: Configuration,
        random_state: Optional[Union[int, np.random.RandomState]],
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        init_params: Optional[Dict[str, Any]] = None,
        dataset_properties: Dict[str, Any] = {},
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ):
        self.config = config
        if config == 1:
            super().__init__(strategy="uniform")
        else:
            super().__init__(strategy="most_frequent")

        self.random_state = random_state
        self.init_params = init_params
        self.dataset_properties = dataset_properties
        self.include = include
        self.exclude = exclude
        self.feat_type = feat_type

    def pre_transform(
        self,
        X: np.ndarray,
        y: np.ndarray,
        fit_params: Optional[Dict[str, Any]] = None,
    ) -> Tuple[np.ndarray, Dict[str, Any]]:
        if fit_params is None:
            fit_params = {}
        return X, fit_params

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[Union[np.ndarray, List]] = None,
    ) -> DummyClassifier:
        return super(MyDummyClassifier, self).fit(
            np.ones((X.shape[0], 1)), y, sample_weight=sample_weight
        )

    def fit_estimator(
        self,
        X: np.ndarray,
        y: np.ndarray,
        fit_params: Optional[Dict[str, Any]] = None,
    ) -> DummyClassifier:
        return self.fit(X, y)

    def predict_proba(self, X: np.ndarray, batch_size: int = 1000) -> np.ndarray:
        new_X = np.ones((X.shape[0], 1))
        probas = super().predict_proba(new_X)
        probas = convert_multioutput_multiclass_to_multilabel(probas).astype(np.float32)
        return probas

    def estimator_supports_iterative_fit(self) -> bool:
        return False

    def get_additional_run_info(self) -> Optional[TYPE_ADDITIONAL_INFO]:
        return None


class MyDummyRegressor(DummyRegressor):
    def __init__(
        self,
        config: Configuration,
        random_state: Optional[Union[int, np.random.RandomState]],
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        init_params: Optional[Dict[str, Any]] = None,
        dataset_properties: Dict[str, Any] = {},
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ):
        self.config = config
        if config == 1:
            super().__init__(strategy="mean")
        else:
            super().__init__(strategy="median")
        self.random_state = random_state
        self.init_params = init_params
        self.dataset_properties = dataset_properties
        self.include = include
        self.exclude = exclude
        self.feat_type = feat_type

    def pre_transform(
        self,
        X: np.ndarray,
        y: np.ndarray,
        fit_params: Optional[Dict[str, Any]] = None,
    ) -> Tuple[np.ndarray, Dict[str, Any]]:
        if fit_params is None:
            fit_params = {}
        return X, fit_params

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[Union[np.ndarray, List]] = None,
    ) -> DummyRegressor:
        return super().fit(np.ones((X.shape[0], 1)), y, sample_weight=sample_weight)

    def fit_estimator(
        self,
        X: np.ndarray,
        y: np.ndarray,
        fit_params: Optional[Dict[str, Any]] = None,
    ) -> DummyRegressor:
        return self.fit(X, y)

    def predict(self, X: np.ndarray, batch_size: int = 1000) -> np.ndarray:
        new_X = np.ones((X.shape[0], 1))
        return super().predict(new_X).astype(np.float32)

    def estimator_supports_iterative_fit(self) -> bool:
        return False

    def get_additional_run_info(self) -> Optional[TYPE_ADDITIONAL_INFO]:
        return None


def _fit_and_suppress_warnings(
    logger: Union[logging.Logger, PicklableClientLogger],
    model: BaseEstimator,
    X: np.ndarray,
    y: np.ndarray,
) -> BaseEstimator:
    def send_warnings_to_log(
        message: Union[Warning, str],
        category: Type[Warning],
        filename: str,
        lineno: int,
        file: Optional[TextIO] = None,
        line: Optional[str] = None,
    ) -> None:
        logger.debug("%s:%s: %s:%s" % (filename, lineno, str(category), message))
        return

    with warnings.catch_warnings():
        warnings.showwarning = send_warnings_to_log
        model.fit(X, y)

    return model


class AbstractEvaluator(object):
    def __init__(
        self,
        backend: Backend,
        queue: multiprocessing.Queue,
        metrics: Sequence[Scorer],
        additional_components: Dict[str, ThirdPartyComponents],
        port: Optional[int],
        configuration: Optional[Union[int, Configuration]] = None,
        scoring_functions: Optional[List[Scorer]] = None,
        seed: int = 1,
        output_y_hat_optimization: bool = True,
        num_run: Optional[int] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        disable_file_output: Union[bool, List[str]] = False,
        init_params: Optional[Dict[str, Any]] = None,
        budget: Optional[float] = None,
        budget_type: Optional[str] = None,
    ):

        # Limit the number of threads that numpy uses
        threadpool_limits(limits=1)

        self.starttime = time.time()

        self.configuration = configuration
        self.backend = backend
        self.port = port
        self.queue = queue

        self.datamanager = self.backend.load_datamanager()
        self.feat_type = self.datamanager.feat_type
        self.include = include
        self.exclude = exclude

        self.X_test = self.datamanager.data.get("X_test")
        self.y_test = self.datamanager.data.get("Y_test")

        self.metrics = metrics
        self.task_type = self.datamanager.info["task"]
        self.seed = seed

        self.output_y_hat_optimization = output_y_hat_optimization
        self.scoring_functions = scoring_functions if scoring_functions else []

        if isinstance(disable_file_output, (bool, list)):
            self.disable_file_output: Union[bool, List[str]] = disable_file_output
        else:
            raise ValueError("disable_file_output should be either a bool or a list")

        if self.task_type in REGRESSION_TASKS:
            if not isinstance(self.configuration, Configuration):
                self.model_class = MyDummyRegressor
            else:
                self.model_class = (
                    autosklearn.pipeline.regression.SimpleRegressionPipeline
                )
            self.predict_function = self._predict_regression
        else:
            if not isinstance(self.configuration, Configuration):
                self.model_class = MyDummyClassifier
            else:
                self.model_class = (
                    autosklearn.pipeline.classification.SimpleClassificationPipeline
                )
            self.predict_function = self._predict_proba

        self._init_params = {"data_preprocessor:feat_type": self.datamanager.feat_type}

        if init_params is not None:
            self._init_params.update(init_params)

        if num_run is None:
            num_run = 0
        self.num_run = num_run

        logger_name = "%s(%d):%s" % (
            self.__class__.__name__.split(".")[-1],
            self.seed,
            self.datamanager.name,
        )

        if self.port is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = get_named_client_logger(
                name=logger_name,
                port=self.port,
            )

        self.X_optimization: Optional[SUPPORTED_XDATA_TYPES] = None
        self.Y_optimization: Optional[SUPPORTED_TARGET_TYPES] = None
        self.Y_actual_train = None

        self.budget = budget
        self.budget_type = budget_type

        # Add 3rd-party components to the list of 3rd-party components in case this
        # wasn't done before (this happens if we run in parallel and the components
        # are only passed to the AbstractEvaluator via the TAE and are not there
        # yet because the worker is in its own process).
        for key in additional_components:
            for component_name, component in additional_components[
                key
            ].components.items():
                if component_name not in _addons[key].components:
                    _addons[key].add_component(component)

        # Please mypy to prevent not defined attr
        self.model = self._get_model(feat_type=self.feat_type)

    def _get_model(self, feat_type: Optional[FEAT_TYPE_TYPE]) -> BaseEstimator:
        if not isinstance(self.configuration, Configuration):
            model = self.model_class(
                feat_type=feat_type,
                config=self.configuration,
                random_state=self.seed,
                init_params=self._init_params,
            )
        else:
            if self.task_type in REGRESSION_TASKS:
                dataset_properties = {
                    "task": self.task_type,
                    "sparse": self.datamanager.info["is_sparse"] == 1,
                    "multioutput": self.task_type == MULTIOUTPUT_REGRESSION,
                }
            else:
                dataset_properties = {
                    "task": self.task_type,
                    "sparse": self.datamanager.info["is_sparse"] == 1,
                    "multilabel": self.task_type == MULTILABEL_CLASSIFICATION,
                    "multiclass": self.task_type == MULTICLASS_CLASSIFICATION,
                }
            model = self.model_class(
                feat_type=feat_type,
                config=self.configuration,
                dataset_properties=dataset_properties,
                random_state=self.seed,
                include=self.include,
                exclude=self.exclude,
                init_params=self._init_params,
            )
        return model

    def _loss(
        self,
        y_true: np.ndarray,
        y_hat: np.ndarray,
        X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
    ) -> Dict[str, float]:
        """Auto-sklearn follows a minimization goal.
        The calculate_loss internally translate a score function to
        a minimization problem.

        For a dummy prediction, the worst result is assumed.

        Parameters
        ----------
            y_true
        """
        return calculate_losses(
            y_true,
            y_hat,
            self.task_type,
            self.metrics,
            X_data=X_data,
            scoring_functions=self.scoring_functions,
        )

    def finish_up(
        self,
        loss: Union[Dict[str, float], float],
        train_loss: Optional[Dict[str, float]],
        opt_pred: np.ndarray,
        test_pred: np.ndarray,
        additional_run_info: Optional[TYPE_ADDITIONAL_INFO],
        file_output: bool,
        final_call: bool,
        status: StatusType,
    ) -> Tuple[
        float,
        Union[float, Dict[str, float]],
        int,
        Dict[str, Union[str, int, float, Dict, List, Tuple]],
    ]:
        """Do everything necessary after the fitting is done:

        * predicting
        * saving the files for the ensembles_statistics
        * generate output for SMAC
        We use it as the signal handler so we can recycle the code for the
        normal usecase and when the runsolver kills us here :)
        """
        self.duration = time.time() - self.starttime

        if file_output:
            file_out_loss, additional_run_info_ = self.file_output(opt_pred, test_pred)
        else:
            file_out_loss = None
            additional_run_info_ = {}

        test_loss = self.calculate_auxiliary_losses(test_pred)

        if file_out_loss is not None:
            return self.duration, file_out_loss, self.seed, additional_run_info_

        loss_ = loss
        for metric in self.metrics:
            if metric.name not in loss_:
                raise ValueError(
                    f"Unable to compute optimization metric {metric.name}. Are you "
                    f"sure {metric.name} is applicable for the given task type?"
                )
        if len(self.metrics) == 1:
            loss = loss_[self.metrics[0].name]
        else:
            loss = {metric.name: loss_[metric.name] for metric in self.metrics}

        additional_run_info = {} if additional_run_info is None else additional_run_info
        for metric in self.scoring_functions:
            if metric.name in loss_:
                additional_run_info[metric.name] = loss_[metric.name]
        additional_run_info["duration"] = self.duration
        additional_run_info["num_run"] = self.num_run
        if train_loss is not None:
            if len(self.metrics) == 1:
                additional_run_info["train_loss"] = train_loss[self.metrics[0].name]
            else:
                additional_run_info["train_loss"] = [
                    train_loss[metric.name] for metric in self.metrics
                ]
        if test_loss is not None:
            additional_run_info["test_loss"] = test_loss

        return_value_dict = {
            "loss": loss,
            "additional_run_info": additional_run_info,
            "status": status,
        }
        if final_call:
            return_value_dict["final_queue_element"] = True

        self.queue.put(return_value_dict)
        return self.duration, loss_, self.seed, additional_run_info_

    def calculate_auxiliary_losses(
        self,
        Y_test_pred: np.ndarray | None,
    ) -> float | dict[str, float] | None:
        if Y_test_pred is None or self.y_test is None:
            return None

        test_loss = self._loss(self.y_test, Y_test_pred)
        if len(self.metrics) == 1:
            test_loss = test_loss[self.metrics[0].name]

        return test_loss

    def file_output(
        self,
        Y_optimization_pred: np.ndarray,
        Y_test_pred: np.ndarray,
    ) -> tuple[float | None, dict[str, Any]]:
        # Abort if self.Y_optimization is None
        # self.Y_optimization can be None if we use partial-cv, then,
        # obviously no output should be saved.
        if self.Y_optimization is None:
            return None, {}

        # Abort in case of shape misalignment
        if np.shape(self.Y_optimization)[0] != Y_optimization_pred.shape[0]:
            return (
                1.0,
                {
                    "error": "Targets %s and prediction %s don't have "
                    "the same length. Probably training didn't "
                    "finish"
                    % (np.shape(self.Y_optimization), Y_optimization_pred.shape)
                },
            )

        # Abort if predictions contain NaNs
        for y, s in [(Y_optimization_pred, "optimization"), (Y_test_pred, "test")]:
            if y is not None and not np.all(np.isfinite(y)):
                return (
                    1.0,
                    {"error": "Model predictions for %s set contains NaNs." % s},
                )

        # Abort if we don't want to output anything.
        # Since disable_file_output can also be a list, we have to explicitly
        # compare it with True.
        if self.disable_file_output is True:
            return None, {}

        # Notice that disable_file_output==False and disable_file_output==[]
        # means the same thing here.
        if self.disable_file_output is False:
            self.disable_file_output = []

        # Here onwards, the self.disable_file_output can be treated as a list
        self.disable_file_output = cast(List, self.disable_file_output)

        # This file can be written independently of the others down bellow
        if "y_optimization" not in self.disable_file_output:
            if self.output_y_hat_optimization:
                self.backend.save_additional_data(
                    self.Y_optimization, what="targets_ensemble"
                )
                self.backend.save_additional_data(
                    self.X_optimization, what="input_ensemble"
                )

        models: Optional[BaseEstimator] = None
        if hasattr(self, "models"):
            if len(self.models) > 0 and self.models[0] is not None:
                if "models" not in self.disable_file_output:

                    if self.task_type in CLASSIFICATION_TASKS:
                        models = VotingClassifier(
                            estimators=None,
                            voting="soft",
                        )
                    else:
                        models = VotingRegressor(estimators=None)
                    # Mypy cannot understand hasattr yet
                    models.estimators_ = self.models  # type: ignore[attr-defined]

        self.backend.save_numrun_to_dir(
            seed=self.seed,
            idx=self.num_run,
            budget=self.budget,
            model=self.model if "model" not in self.disable_file_output else None,
            cv_model=models if "cv_model" not in self.disable_file_output else None,
            # TODO: below line needs to be deleted once backend is updated
            valid_predictions=None,
            ensemble_predictions=(
                Y_optimization_pred
                if "y_optimization" not in self.disable_file_output
                else None
            ),
            test_predictions=(
                Y_test_pred if "y_test" not in self.disable_file_output else None
            ),
        )

        return None, {}

    def _predict_proba(
        self,
        X: np.ndarray,
        model: BaseEstimator,
        task_type: int,
        Y_train: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        def send_warnings_to_log(
            message: Union[Warning, str],
            category: Type[Warning],
            filename: str,
            lineno: int,
            file: Optional[TextIO] = None,
            line: Optional[str] = None,
        ) -> None:
            self.logger.debug(
                "%s:%s: %s:%s" % (filename, lineno, str(category), message)
            )
            return

        with warnings.catch_warnings():
            warnings.showwarning = send_warnings_to_log
            Y_pred = model.predict_proba(X, batch_size=1000)

        if Y_train is None:
            raise ValueError("Y_train is required for classification problems")

        Y_pred = self._ensure_prediction_array_sizes(Y_pred, Y_train)
        return Y_pred

    def _predict_regression(
        self,
        X: np.ndarray,
        model: BaseEstimator,
        task_type: int,
        Y_train: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        def send_warnings_to_log(
            message: Union[Warning, str],
            category: Type[Warning],
            filename: str,
            lineno: int,
            file: Optional[TextIO] = None,
            line: Optional[str] = None,
        ) -> None:
            self.logger.debug(
                "%s:%s: %s:%s" % (filename, lineno, str(category), message)
            )
            return

        with warnings.catch_warnings():
            warnings.showwarning = send_warnings_to_log
            Y_pred = model.predict(X)

        if len(Y_pred.shape) == 1:
            Y_pred = Y_pred.reshape((-1, 1))

        return Y_pred

    def _ensure_prediction_array_sizes(
        self, prediction: np.ndarray, Y_train: np.ndarray
    ) -> np.ndarray:
        num_classes = self.datamanager.info["label_num"]

        if (
            self.task_type == MULTICLASS_CLASSIFICATION
            and prediction.shape[1] < num_classes
        ):
            if Y_train is None:
                raise ValueError("Y_train must not be None!")
            classes = list(np.unique(Y_train))

            mapping = dict()
            for class_number in range(num_classes):
                if class_number in classes:
                    index = classes.index(class_number)
                    mapping[index] = class_number
            new_predictions = np.zeros(
                (prediction.shape[0], num_classes), dtype=np.float32
            )

            for index in mapping:
                class_index = mapping[index]
                new_predictions[:, class_index] = prediction[:, index]

            return new_predictions

        return prediction



from typing import Any, Dict, List, Optional, Tuple, Union

import multiprocessing
import queue

__all__ = ["read_queue"]


def read_queue(
    queue_: multiprocessing.Queue,
) -> List[Dict[str, Union[str, bool, int, float, List, Dict, Tuple]]]:
    stack = []
    while True:
        try:
            return_value = queue_.get(timeout=1)
        except queue.Empty:
            break

        # Check if there is a special placeholder value which tells us that
        # we don't have to wait until the queue times out in order to
        # retrieve the final value!
        if "final_queue_element" in return_value:
            del return_value["final_queue_element"]
            do_break = True
        else:
            do_break = False
        stack.append(return_value)
        if do_break:
            break

    if len(stack) == 0:
        raise queue.Empty
    else:
        return stack


def empty_queue(queue_: multiprocessing.Queue) -> None:
    while True:
        try:
            queue_.get(block=False)
        except queue.Empty:
            break

    queue_.close()


def extract_learning_curve(
    stack: List[Dict[str, Any]], key: Optional[str] = None
) -> List[float]:
    learning_curve = []
    for entry in stack:
        if key:
            learning_curve.append(entry["additional_run_info"][key])
        else:
            learning_curve.append(entry["loss"])
    return list(learning_curve)
